#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import ipaddress
import os
import re
import subprocess
import sys

import ucpeif.loopif as loopif
import ucpeif.pppoeclientif as pppoeif
import vppapi_if.vppapi_if as vppif
from pydbus import SystemBus
from pyroute2 import IPDB
from sysrepo_api import sysrepo_api3 as api
from ucpeif import ipsecif as ips
from time import sleep
from sysrepo_interfaces import _sysrepo_interfaces_wifi as wifi
from sysrepo_interfaces import radvd_configuration as radvd
from sysrepo_interfaces import utils as utils
from sysrepo_interfaces import vpp as vpputils
from sysrepo_interfaces import nat as natutils
from sysrepo_interfaces import wireguard as wiregd
from sysrepo_interfaces import ipsecvpn as ipsec
from sysrepo_interfaces import openvpn as openvpn
from sysrepo_interfaces import vxlan as vxlan
from sysrepo_interfaces import vlan as vlanutils

svc_restart_file = "/config/svcs_to_restart"

sys.path.append('/scripts')

TVEAPI_PORT = 7373

class IfUpError(Exception):
    def __init__(self, message, exit_code):
        super(IfUpError, self).__init__(message)
        self.exit_code = exit_code


class SysrepoInterfaces(api.SysrepoAPI):
    VXLAN_OFFSET = 500
    BR_PREFFIX = 'br-'

    def __init__(self,
                 app_name="interfaces",
                 logger=None,
                 log_file=None,
                 log_level=None,
                 console_log=False,
                 journal_log=True,
                 indexes={'': 'name', 'wan-interface': 'name', 'proposals': 'id', 'dh-groups': 'group', 'rounds': 'id',
                          'wifi-wan-interface': 'name', 'mac-allow': 'mac', 'mac-deny': 'mac', 'probes': 'id',
                          'peers': 'peer-endpoint', 'smart-protocol-passthrough': 'protocol'},
                 namespace="/ucpe:config/interfaces:interface",
                 extra_ns=['/ucpe:config/system:system'],
                 status_namespace="/ucpe:status/interfaces:interface",
                 run_as_subscriber=False,
                 priority=0):
        api.SysrepoAPI.__init__(
            self,
            app_name=app_name,
            logger=logger,
            log_file=log_file,
            log_level=log_level,
            console_log=console_log,
            journal_log=journal_log,
            namespace=namespace,
            extra_ns=extra_ns,
            status_namespace=status_namespace,
            indexes=indexes,
            run_as_subscriber=run_as_subscriber,
            priority=priority
        )
        self.bus = None
        self.natpassthr = None
        self.tracehost = None
        self.ipsvpn_interfaces = []
        self.vxlan_interfaces = []
        self.pppoe_interfaces = []
        self.loopback_interfaces = []
        self.wg_interfaces = []
        self.hw_type = utils.get_hw_type().lower()
        self.radvd_ifs = {}
        self.pciids = {}
        self.igb_vfs = {}
        self.initramfs_changed = False
        self.igb_recalc_needed = False

    def run_subscriber(self):
        try:
            self.bus = SystemBus()
            if self.bus is not None:
                self.tracehost = self.bus.get('net.telco.TracehostMonitor')
        except:
            self.logger.error('Connecting to \'net.telco.TracehostMonitor\' failed !')

        if not api.is_nanopi():
            try:
                self.bus = SystemBus()
                if self.bus is not None:
                    self.natpassthr = self.bus.get('net.telco.NatIfChange')
            except:
                self.logger.error('Connecting to \'net.telco.NatIfChange\' failed !')

        utils.init_pci_data(self)
        api.SysrepoAPI.run_subscriber(self)

    def get_if_sysname(self, name):
        return utils.get_ifname(self, name)

    def get_ip_mask_kernel_intfc(self, intfc):
        with IPDB() as ip:
            iface = ip.interfaces[intfc]
            ip_address = iface.ipaddr[0]['address']
            netmask = iface.ipaddr[0]['prefixlen']
            return ip_address, netmask

    def __config_mgmt_interface(self, name, config, change_set):

        bootproto = self.get_interface_param(name, 'bootproto', 'dhcp')
        ipv4_prefix = self.get_interface_param(name, 'ipv4-prefix', '')
        ipv6_prefix = self.get_interface_param(name, 'ipv6-prefix', '')
        gateway = self.get_interface_param(name, 'gateway', '')
        gateway6 = self.get_interface_param(name, 'gateway6', '')
        metric = self.get_interface_param(name, 'metric', 100)
        management = self.get_interface_param(name, 'management', '')
        management_old = 'false'

        if change_set:
            for xpath, change in change_set.items():
                if change['operation'] == 'DELETED':
                    # Fill defaults
                    if xpath.find('bootproto') > -1:
                        bootproto = ''
                    if xpath.find('ipv4-prefix') > -1:
                        ipv4_prefix = ''
                    if xpath.find('ipv6-prefix') > -1:
                        ipv6_prefix = ''
                    if xpath.find('dns') > -1:
                        dns = ''
                    if xpath.find('dns6') > -1:
                        dns6 = ''
                    if xpath.find('metric') > -1:
                        metric = 100
                    if xpath.find('vlan') > -1:
                        vlan = 0
                    if xpath.find('gateway6') > -1:
                        gateway6 = ''
                    elif xpath.find('gateway') > -1:
                        gateway = ''
                    if xpath.find('management') > -1:
                        management = 'false'
                    continue
                if xpath.find('ipv4-prefix') > -1:
                    ipv4_prefix = change['new_value']
                if xpath.find('ipv6-prefix') > -1:
                    ipv6_prefix = change['new_value']
                if xpath.find('gateway6') > -1:
                    gateway6 = change['new_value']
                elif xpath.find('gateway') > -1:
                    gateway = change['new_value']
                if xpath.find('management') > -1:
                    management = change['new_value']
                    if 'old_value' in change:
                        management_old = change['old_value']
                if xpath.find('metric') > -1:
                    metric = change['new_value']

        if ipv4_prefix or ipv6_prefix:
            bootproto = "static"
        else:
            bootproto = "manual"

        utils.shell_return_ok(self,"ifdown --ignore-error --force {}".format(name))
        utils.shell_return_ok(self,"rm -rf /etc/network/interfaces.d/ifcfg-{}".format(name))
        onbt = "yes"
        options = ["-n {}".format(name), "-o {}".format(onbt)]
        options.append("-b {}".format(bootproto))
        if ipv4_prefix:
            options.append("--ipv4-prefix={}".format(ipv4_prefix))
        if ipv6_prefix:
            options.append("--ipv6-prefix={}".format(ipv6_prefix))
        if gateway:
            options.append("-g {}".format(gateway))
        if gateway6:
            options.append("--ipv6-gateway={}".format(gateway6))
            #options.append('-j"ip -6 route add {gateway} dev {interface} && ip -6 route add default via {gateway}"'.format(gateway=gateway6, interface=ifname))
            options.append('-k"ip -6 route del {gateway} dev {interface} && ip -6 route del default via {gateway}"'.format(gateway=gateway6, interface=ifname))
        if metric:
            options.append("-c {}".format(metric))

        options.append("--ovs-master ovs --ovs-ports {}".format(name))
        self.execute_shell_command("/scripts/apply_networking.py {}".format(" ".join(options)))

        # Configure and enable RA for IPv6
        if ipv6_prefix:
            self.radvd_ifs[name] = ipv6_prefix
        else:
            try:
                if name in self.radvd_ifs.keys():
                    del self.radvd_ifs[name]
            except:
                self.logger.info(f'Cannot delete radvd config for: {name}')

        if onbt == "yes":
            try:
                self.logger.info("ifup interface {}".format(name))
                subprocess.check_call(['ifup', '--ignore-error', name])
                if str(self.execute_shell_command("systemctl is-active isc-dhcp-relay")).strip("\n") == "active":
                    if self.get_item("/ucpe:config/dhcprelay:dhcp-relay/dhcp-relay-interface") == "mgmt0":
                        self.execute_shell_command("systemctl --no-block restart isc-dhcp-relay.service")
            except subprocess.CalledProcessError as ex:
                self.logger.error('ifup command failed!')
                utils.shell_return_ok(self,"ifdown --ignore-error --force {}".format(name))
                raise IfUpError('ifup error: {}'.format(ex), 1)

    def __config_interface_loopback(self, name, config, change_set):
        self.logger.debug("__config_interface_loopback")
        self.logger.debug("Config {}".format(config))
        self.logger.debug("Change set {}".format(change_set))

        # Getting network-mode status
        mode = "VPP" if self.get_network_mode() == 'vpp' else "OVS"
        self.logger.debug(f"Network mode: {mode}")

        if not change_set and config:
            # Startup config
            loop_if = next((a for a in self.loopback_interfaces if a.name == config.get('name')), None)
            if loop_if:
                self.logger.debug("Interface \'{}\' already exists.".format(loop_if.name))
            else:
                ipv4 = config.get('ipv4-prefix', None)
                if ipv4:
                    ipv4 = ipaddress.ip_interface(ipv4)

                ipv6 = config.get('ipv6-prefix', None)
                if ipv6:
                    ipv6 = ipaddress.ip_interface(ipv6)

                enable = config.get('enable', False)

                loop_if = loopif.LoopInterface (config.get('name'), ipv4_prefix=ipv4, ipv6_prefix=ipv6, enable=enable, logger=self.logger, network_mode=mode)
                self.loopback_interfaces.append(loop_if)
            return

        if change_set:
            ch_set_keys = change_set.keys()

            self.logger.info('Keys : {}'.format(ch_set_keys))
            loop_if = next((loop for loop in self.loopback_interfaces if loop.name == name), None)
            if 'name' in ch_set_keys:
                if change_set.get('name', None):
                    if change_set.get('name').get('operation', None) == 'CREATED':
                        if loop_if:
                            self.logger.error("There is object for {} ?!?!".format(name))
                        else:
                            loop_if = loopif.LoopInterface(name, logger=self.logger, network_mode=mode)
                            self.loopback_interfaces.append(loop_if)
                    elif change_set.get('name').get('operation', None) == 'DELETED':
                        if loop_if:
                            self.loopback_interfaces.remove(loop_if)
                            del loop_if
                        return

            if 'enable' in ch_set_keys:
                oper = change_set.get('enable').get('operation', None)
                if oper == 'MODIFIED' or oper == 'CREATED':
                    enable = change_set.get('enable').get('new_value')
                    enable = True if 'true' in str(enable).lower() else False
                else:
                    enable = False

                self.logger.debug(str(loop_if))
                loop_if.enable = enable

            if 'ipv4-prefix' in ch_set_keys:
                oper = change_set.get('ipv4-prefix').get('operation', None)
                if oper == 'MODIFIED' or oper == 'CREATED':
                    ipv4_pr = change_set.get('ipv4-prefix').get('new_value')
                    loop_if.ip4 = ipaddress.IPv4Interface(str(ipv4_pr))
                else:
                    loop_if.ip4 = None
                loop_if.enable = False
                loop_if.enable = True


            if 'ipv6-prefix' in ch_set_keys:
                oper = change_set.get('ipv6-prefix').get('operation', None)
                if oper == 'MODIFIED' or oper == 'CREATED':
                    ipv6_pr = change_set.get('ipv6-prefix').get('new_value')
                    loop_if.ip6 = ipaddress.IPv6Interface(str(ipv6_pr))
                else:
                    loop_if.ip6 = None
                loop_if.enable = False
                loop_if.enable = True

            for key in change_set.keys():
                self.logger.debug('Key \'{}\' operation {}'.format(key, change_set.get(key).get('operation')))

    def __config_interface_pppoe(self, name, config, change_set):
        if self.get_network_mode() != 'vpp':
            self.logger.error('uCPE is not in VPP mode')
            return

        c_auth_type = pppoeif.AUTH_TYPE.PAP
        c_username = None
        c_password = None
        c_ethif = None
        c_status = False
        c_idle = 0
        c_retry = 0
        c_metric = 130
        c_management = False
        c_ac_name = None
        c_ippass = False

        self.logger.debug("Config {}".format(config))
        if config:
            if config.get('auth-type', 'pap') == 'chap':
                c_auth_type = pppoeif.AUTH_TYPE.CHAP
            
            c_username = config.get('username', None)
            c_password = config.get('password', None)
            c_ethif = config.get('ethernet-interface', None)
            c_status = config.get('enable', False)
            c_idle = config.get('idle-timeout', 0)
            c_retry = config.get('retry-timeout', 0)
            c_metric = config.get('metric', 130)
            c_management = config.get('management', False)
            c_ac_name = config.get('ac-name', None)
            c_ipv6 = config.get('ipv6', False)
            c_ippass = config.get('ip-passthrough', False)

            ifname = utils.get_ifname(self,name)

            #To-Do get ethernet-interface mode
            if c_ethif:
                if 'virtio' == self.get_interface_param(c_ethif, 'mode', 'virtio'):
                    c_ethif = 'host-{}'.format(c_ethif)
                c_ethif = self.tag_interface(c_ethif,  self.get_interface_param(c_ethif.replace('host-', ''), 'vlan', None))

        if not change_set and config:
            # Startup config
            self.logger.debug("APPLYING STARTUP FOR PPPOE INTERFACE - {}".format(name))
            ppp_if = next(( a for a in self.pppoe_interfaces if a.name == config.get('name')), None)
            if ppp_if:
                self.logger.debug("Interface \'{}\' allready exists.".format(ppp_if.name))
            else:
                self.logger.debug("Add PPPoE configuration for \'{}\' interface" .format(config.get('name')))
                ppp_if = pppoeif.PPPoEClientInterface(config.get('name'), auth_type=c_auth_type, username=c_username, password=c_password, 
                                                        ethif=c_ethif, logger=self.logger, enable=c_status, metric=c_metric,
                                                        idle_timeout=c_idle, retry_timeout=c_retry, ipv6=c_ipv6, ac_name=c_ac_name,
                                                        ip_pass=c_ippass)

                self.pppoe_interfaces.append(ppp_if)

                mtu = config.get('mtu', 1492)
                if mtu:
                    ppp_if.mtu = int(mtu)

                utils.apply_management_rules(self, ifname, False, c_management, change_set)

        if change_set:
            self.logger.debug("APPLYING RUNNING  FOR PPOE INTERFACE - {}".format(name))
            ch_set_keys = change_set.keys()
            self.logger.debug('Keys : {}'.format(ch_set_keys))
            # n_status = c_status

            ppp_if = next(( a for a in self.pppoe_interfaces if a.name == name), None)

            if 'name' in ch_set_keys:
                if change_set.get('name', None):
                    if change_set.get('name').get('operation', None) == 'CREATED':
                        if ppp_if:
                            self.logger.error("There is PPPoE interface object for {} ?!?!".format(name))
                        else:
                            ppp_if = pppoeif.PPPoEClientInterface( name, logger=self.logger )
                            self.pppoe_interfaces.append(ppp_if)
                    elif change_set.get('name').get('operation', None) == 'DELETED':
                        if ppp_if:
                            self.pppoe_interfaces.remove(ppp_if)
                            del ppp_if
                        return

            # # if current status is enable - disable PPPoE connection (in order to apply all new values) and if new status is enable - reconnect at the end
            # if c_status is True:
            #     ppp_if.enable = False

            if 'auth-type' in ch_set_keys:
                oper = change_set.get('auth-type').get('operation', None )
                n_auth_type = None

                self.logger.debug('Auth type is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    if change_set.get('auth-type').get('new_value', 'pap') == 'pap':
                        n_auth_type = pppoeif.AUTH_TYPE.PAP
                    elif change_set.get('auth-type').get('new_value', 'pap') == 'chap':
                        n_auth_type = pppoeif.AUTH_TYPE.CHAP

                ppp_if.auth_type = n_auth_type

            if 'ethernet-interface' in ch_set_keys:
                oper = change_set.get('ethernet-interface').get('operation', None )
                n_ethif = None

                self.logger.debug('Eth if is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_ethif = change_set.get('ethernet-interface').get('new_value', None)

                if n_ethif:
                    if 'virtio' == self.get_interface_param(n_ethif, 'mode', 'virtio'):
                        n_ethif = 'host-{}'.format(n_ethif)
                    if 'eth-if-vlan' in ch_set_keys:
                        vlan = change_set.get('eth-if-vlan').get('new_value')
                        n_ethif += f'.{vlan}'
                    else:
                        n_ethif = self.tag_interface(n_ethif, self.get_interface_param(n_ethif.replace('host-', ''), 'vlan',
                                                                                   None))
                ppp_if.ethif = n_ethif

            if 'username' in ch_set_keys:
                oper = change_set.get('username').get('operation', None )
                n_username = None

                self.logger.debug('Username is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_username = change_set.get('username').get('new_value', None)

                ppp_if.username = n_username

            if 'password' in ch_set_keys:
                oper = change_set.get('password').get('operation', None )
                n_password = None

                self.logger.debug('Password is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_password = change_set.get('password').get('new_value', None)

                ppp_if.password = n_password

            if 'metric' in ch_set_keys:
                oper = change_set.get('metric').get('operation', None )
                n_metric = 130

                self.logger.debug('Metric is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_metric = change_set.get('metric').get('new_value', 130)

                ppp_if.metric = int(n_metric)
                if oper == 'MODIFIED' and c_status is not False:
                    ipdb = IPDB()
                    remote_ip = ''
                    local_ip = ''

                    if name in ipdb.interfaces:
                        ppp_idx = ipdb.interfaces[name]
                        for rt in ipdb.routes:
                            if rt['family'] == 10:
                                continue
                            if rt['oif'] == ppp_idx['index']:
                                if rt['prefsrc']:
                                    local_ip = rt['prefsrc']
                                if rt['dst']:
                                    remote_ip = rt['dst'].split('/')[0]

                    try:
                        cmd = ['/opt/bin/pppif_host_cfg.py', '-n', name, '-l', local_ip, '-r', remote_ip]
                        subprocess.check_call(cmd)
                    except subprocess.CalledProcessError as e:
                        self.logger.debug(e.output)


            if 'idle-timeout' in ch_set_keys:
                oper = change_set.get('idle-timeout').get('operation', None )
                n_idle_timeout = 0

                self.logger.debug('Idle timeout is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_idle_timeout = change_set.get('idle-timeout').get('new_value', 0)

                ppp_if.idle_timeout = int(n_idle_timeout)

            if 'retry-timeout' in ch_set_keys:
                oper = change_set.get('retry-timeout').get('operation', None )
                n_retry_timeout = 0

                self.logger.debug('Retry timeout is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_retry_timeout = change_set.get('retry-timeout').get('new_value', 0)

                ppp_if.retry_timeout = int(n_retry_timeout)

            if 'mtu' in ch_set_keys:
                oper = change_set.get('mtu').get('operation', None )
                n_mtu = 1492

                self.logger.debug('MTU is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_mtu = change_set.get('mtu').get('new_value', 1492)

                ppp_if.mtu = int(n_mtu)

            if 'enable' in ch_set_keys:
                oper = change_set.get('enable').get('operation', None)
                n_status = False

                self.logger.debug('State is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    if 'true' in str(change_set.get('enable').get('new_value', False)).lower():
                        n_status = True
                    
                ppp_if.enable = n_status

            if 'management' in ch_set_keys:
                oper = change_set.get('management').get('operation', None)
                n_management = False

                self.logger.debug('Management is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    if 'true' in str(change_set.get('management').get('new_value', False)).lower():
                        n_management = True

                utils.apply_management_rules(self, name, c_management, n_management, change_set)

            if 'ipv6' in ch_set_keys:
                oper = change_set.get('ipv6').get('operation', None)
                n_ipv6 = False

                self.logger.debug('IPv6 is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    if 'true' in str(change_set.get('ipv6').get('new_value', False)).lower():
                        n_ipv6 = True
                    
                ppp_if.ipv6 = n_ipv6

            if 'ac-name' in ch_set_keys:
                oper = change_set.get('ac-name').get('operation', None )
                n_ac_name = None

                self.logger.debug('AC-name is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_ac_name = change_set.get('ac-name').get('new_value', None)

                ppp_if.ac_name = n_ac_name

    def __vswitch_if_config(self, if_name, config, change_set, all_change_set):
        interface_type = config.get('type', None)
        if interface_type not in ['ethernet', 'vlan'] or if_name == 'mgmt0':
            return

        if_def_vsw = 'br-' + if_name
        if_vsw = config.get('vswitch', None)
        if_mode = config.get('mode', 'virtio')
        
        if change_set:
            self.logger.debug("APPLYING RUNNING VSWITCH FOR INTERFACE - {}".format(if_name))
            ch_set_keys = change_set.keys()

            if 'vswitch' in ch_set_keys:
                oper = change_set.get('vswitch').get('operation', None )
                
                self.logger.debug('VSwitch is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    if_vsw = change_set.get('vswitch').get('new_value', None)
                    
        if not if_vsw:
            self.logger.error("Missing VSwitch configuration")
            return
        vsw_name = if_vsw.strip("br-")

        if all_change_set:
            ch_set_keys = None
            if vsw_name in all_change_set.keys():
                ch_set_keys = all_change_set.get(vsw_name, None)

        if self.get_network_mode() == 'ovs':
            self.logger.debug('Not implemented')
            return
        else:
            if utils.vpp_not_running():
                self.logger.debug('VPP is not running')
                return
            vpputils.interface_to_vsw_vpp(self, if_name, if_mode, if_vsw)

    def __config_interface(self, name, config, change_set, startup=False):
        enable = self.get_interface_param(name, 'enable', "")
        interface_type = self.get_interface_param(name, 'type', None)
        apn_changed = False
        user_changed = False
        ifmode_changed = False
        
        if not interface_type:
            if change_set:
                type_elem = change_set.get('type', None)
                if type_elem and type_elem.get('operation', None) == 'CREATED':
                    interface_type = type_elem.get('new_value', None)

        if not interface_type:
            self.logger.error('{} no type'.format(name))

        if interface_type == 'openvpn' or 'tun0' in name:
            openvpn.config_openvpn_interface(self, name, config, change_set)
            return

        if interface_type == 'ipsecvpn' or 'vti' in name:
            if change_set:
                ipsec.config_ipsecvpn_apply(self, name, change_set, self.ipsvpn_interfaces)
            else:
                ipsec.config_ipsecvpn_startup(self, name, config, self.ipsvpn_interfaces)
            return

        if interface_type == 'vxlan' and self.get_network_mode() == 'vpp':
            vxlan.config_interface_vxlan(self, name, config, change_set)
            return

        if interface_type == 'pppoe':
            self.__config_interface_pppoe(name, config, change_set)
            return

        if interface_type == 'loopback':
            self.__config_interface_loopback(name, config, change_set)
            return

        if interface_type == 'wireguard':
            wiregd.config_interface_wireguard(self, name, config, change_set)
            return

        if interface_type == 'vlan':
            vlanutils.config_interface_vlan(self, name, config, change_set, startup)
            return

        if_names = [f for f in os.listdir('/sys/class/net') if
                    os.path.isfile(os.path.join('/sys/class/net', f)) or
                    os.path.islink(os.path.join('/sys/class/net', f))]

        basename = re.sub(r'(.*)(\.[0-9]+)$', r'\1', name)  # vlan tag interface - check for base name
        if basename not in if_names \
           and 'dpdk' != self.get_interface_param(name, 'mode', 'virtio') \
               and 'lte' != interface_type \
                   and 'wifi' != interface_type:
            self.logger.info("Interface {} not found in the system! Skipping configuration...".format(name))
            return

        if 'nat' == interface_type or 'nat' in name:
            natutils.config_nat_interface(self, name, config, change_set)
            return

        if 'mgmt' in name:
            self.__config_mgmt_interface(name, config, change_set)
            return

        if not config:
            # Interface is deleted from config
            self.execute_shell_command("/scripts/apply_networking.py -z {}".format(name))
            return

        bootproto = self.get_interface_param(name, 'bootproto', 'dhcp')
        ipv4_prefix = self.get_interface_param(name, 'ipv4-prefix', '')
        ipv6_prefix = self.get_interface_param(name, 'ipv6-prefix', '')
        dns = self.get_interface_param(name, 'dns', '')
        dns6 = self.get_interface_param(name, 'dns6', '')
        gateway = self.get_interface_param(name, 'gateway', '')
        gateway6 = self.get_interface_param(name, 'gateway6', '')
        metric = self.get_interface_param(name, 'metric', 100)
        management = self.get_interface_param(name, 'management', '')
        management_old = 'false'

        ifname_old = utils.get_ifname(self, name)
        if 'inband' in name:
            ifname_old = name

        if change_set:
            for xpath, change in change_set.items():
                if change['operation'] == 'DELETED':
                    # Fill defaults
                    if xpath.find('bootproto') > -1:
                        bootproto = 'dhcp'
                    if xpath.find('ipv4-prefix') > -1:
                        ipv4_prefix = ''
                    if xpath.find('ipv6-prefix') > -1:
                        ipv6_prefix = ''
                    if xpath.find('dns6') > -1:
                        dns6 = ''
                    elif xpath.find('dns') > -1:
                        dns = ''
                    if xpath.find('metric') > -1:
                        metric = 100
                    if xpath.find('gateway6') > -1:
                        gateway6 = ''
                    elif xpath.find('gateway') > -1:
                        gateway = ''
                    if xpath.find('management') > -1:
                        management = 'false'
                    if xpath.find('apn') > -1:
                        config['apn'] = ""
                        apn_changed = True
                    if xpath.find('user') > -1:
                        user_changed = True
                        config['user'] = ""
                    if xpath.find('pass') > -1:
                        user_changed = True
                        config['pass'] = ""
                    continue
                if xpath.find('enable') > -1 and 'tracehost' not in xpath and 'smart-protocol-passthrough' not in xpath:
                    enable = change['new_value']
                if xpath.find('bootproto') > -1:
                    bootproto = change['new_value']
                if xpath.find('ipv4-prefix') > -1:
                    ipv4_prefix = change['new_value']
                if xpath.find('ipv6-prefix') > -1:
                    ipv6_prefix = change['new_value']
                if xpath.find('dns6') > -1:
                    dns6 = change['new_value']
                elif xpath.find('dns') > -1:
                    dns = change['new_value']
                if xpath.find('metric') > -1:
                    metric = change['new_value']
                if xpath.find('gateway6') > -1:
                    gateway6 = change['new_value']
                elif xpath.find('gateway') > -1:
                    gateway = change['new_value']
                if xpath.find('management') > -1:
                    management = change['new_value']
                    if 'old_value' in change:
                        management_old = change['old_value']
                if xpath.find('apn') > -1:
                    apn_changed = True
                    config['apn'] = change['new_value']
                if xpath.find('user') > -1:
                    user_changed = True
                    config['user'] = change['new_value']
                if xpath.find('pass') > -1:
                    user_changed = True
                    config['pass'] = change['new_value']

        if (enable == "false" or enable == "False" or not enable) and interface_type == "lte":
            self.logger.info("Disabling LTE iface")
            utils.shell_return_ok(self,"ifdown --ignore-error --force {}".format(ifname_old))
            self.execute_shell_command("/scripts/apply_networking.py -z {}".format(ifname_old))
            utils.shell_return_ok(self,"rm -rf /etc/network/interfaces.d/ifcfg-{}".format(ifname_old))
            return

        if (enable == "false" or enable == "False" or not enable) and interface_type == "wifi":
            self.logger.info("Disabling wifi iface")
            self.execute_shell_command("systemctl stop wpa_supplicant.service hostapd.service")
            self.execute_shell_command("systemctl disable wpa_supplicant.service hostapd.service")
            utils.shell_return_ok(self,"ifdown --ignore-error --force {}".format(ifname_old))
            self.execute_shell_command("/scripts/apply_networking.py -z {}".format(ifname_old))
            utils.shell_return_ok(self,"rm -rf /etc/network/interfaces.d/ifcfg-{}".format(ifname_old))
            return

        if not ifmode_changed:
            ifmode = self.get_interface_param(name, 'mode', 'virtio')
        if ifmode == 'virtio' or ifmode == 'sr-iov':
            self.logger.info("Interface \'{}\' is in \'{}\' mode. Create minimal ifcfg file".format(name, ifmode))
            options = ["-n {}".format(name), "-o {}".format('yes'), "-b manual", "-p \'sysctl -w net.ipv6.conf.{}.disable_ipv6=1\'".format(name)]
            if bootproto.find("6") < 0 or not ipv6_prefix:
                utils.shell_return_ok(self,"echo 0 > /proc/sys/net/ipv6/conf/{}/autoconf".format(name))
            self.execute_shell_command("/scripts/apply_networking.py {}".format(" ".join(options)))
            utils.do_ifup(self, name)
            utils.set_igb_param(self, name, ifmode)

        if 'lte' == interface_type:
            self.logger.info("Interface \'{}\' is in LTE. Create minimal ifcfg file".format(name))
            options = ["-n {}".format(name), "-o {}".format('yes'), "-b manual"]
            self.execute_shell_command("/scripts/apply_networking.py {}".format(" ".join(options)))
            if (enable == "true" or enable):
                utils.shell_return_ok(self,"ifup --ignore-error {}".format(name))

        if 'wifi' == interface_type:
            self.logger.info("Interface {} is in wifi. Create minimal ifcfg file".format(name))
            # options = ["-n {}".format(name), "-o {}".format('yes'), "-b manual"]
            options = [ \
                        "-n {}".format(name), \
                        "-o {}".format('yes'), \
                        "-b manual", \
                        "-p \'sysctl --ignore -w net.ipv6.conf.{}.disable_ipv6=0\'".format(name), \
                        "-p \'sysctl --ignore -w net.ipv6.conf.br-{}.disable_ipv6=0\'".format(name), \
                       ]
            self.execute_shell_command("/scripts/apply_networking.py {}".format(" ".join(options)))
            utils.shell_return_ok(self,"ifup --ignore-error {}".format(name))
            wifi.config_wifi_interface(self, name, config, change_set)

        ifname = utils.get_ifname(self,name)
        if 'inband' in name:
            ifname = name

        if bootproto not in ['dhcp', 'dhcp6', 'dhcp46', 'manual', 'static']:
            bootproto = None

        if not bootproto:
            if ipv4_prefix or ipv6_prefix:
                bootproto = "static"
            else:
                bootproto = "manual"

        onbt = "yes"
        if 'lte' == interface_type:
            if not os.path.exists("/sys/class/net/{}".format(name)) or (str(enable).lower() != 'true' and config['enable'] == False):
                onbt = "no"
            else:
                onbt = "yes"
        elif 'wifi' == interface_type:
            if not os.path.exists("/sys/class/net/{}".format(name)) or (str(enable).lower() != 'true' and config['enable'] == False):
                onbt = "no"
            else:
                onbt = "yes"
        elif 'ethernet' == interface_type:
            onbt = "yes"

        self.logger.info('{} onbt is {}'.format(name, onbt))

        vpputils.config_interface_ippass(self, name, config, change_set)

        # Reconnect all ipsecvpn that have local IP equal to IP of adaptor that is going to be down. Disconnect first.
        ipsecvpn_reenable_list = ipsec.ipsecvpn_for_reenable(self, ifname_old)
        ipsecvpn_reenable_count = len(ipsecvpn_reenable_list)
        if ipsecvpn_reenable_count > 0:
            self.logger.debug("{} ipsecvpn(s) need to be reenabled!".format(ipsecvpn_reenable_count))
            for ipsec_iface in ipsecvpn_reenable_list:
                ipsec.ipsecvpn_disconnect(self, ipsec_iface)

        utils.shell_return_ok(self,"ifdown --ignore-error --force {}".format(ifname_old))
        utils.shell_return_ok(self,"rm -rf /etc/network/interfaces.d/ifcfg-{}".format(ifname_old))

        if 'lte' == interface_type:
            from tveif import if_lte
            if apn_changed or user_changed:
                utils.shell_return_ok(self,"rm -f /var/lib/dhcp/dhclient.{}.leases".format(ifname_old))
            modem_type=if_lte.lte_get_driver_type(config["name"])
            if modem_type == 1:
                if_lte.reconnect_lte_qmi_dev(config, if_lte.modem_default)
            elif modem_type == 2 or modem_type == 4:
                if_lte.setup_lte_mbim_dev(config,if_lte.mbim_modem_default) 
                

        utils.apply_management_rules(self, ifname, management_old, management, change_set)

        options = ["-n {}".format(ifname), "-o {}".format(onbt)]
        if bootproto:
            options.append("-b {}".format(bootproto))
        if ipv4_prefix:
            options.append("--ipv4-prefix={}".format(ipv4_prefix))
        if ipv6_prefix:
            options.append("--ipv6-prefix={}".format(ipv6_prefix))
            options.append('-p"echo 0 > /proc/sys/net/ipv6/conf/%s/accept_dad"' % str(ifname))
        if dns:
            options.append("-d {}".format(dns))
        if dns6:
            options.append("--ipv6-dns={}".format(dns6))
        if gateway:
            options.append("-g {}".format(gateway))
        if gateway6:
            options.append("--ipv6-gateway={}".format(gateway6))
            #options.append('-j"ip -6 route add {gateway} dev {interface} && ip -6 route add default via {gateway}"'.format(gateway=gateway6, interface=ifname))
            options.append('-k"ip -6 route del {gateway} dev {interface} && ip -6 route del default via {gateway}"'.format(gateway=gateway6, interface=ifname))
        if metric:
            options.append("-c {}".format(metric))
        if ifname.find("br") > -1:
            options.append("--ovs-master ovs --ovs-ports {}".format(name))
            if bootproto.find("6") < 0 or not ipv6_prefix:
                utils.shell_return_ok(self,"echo 0 > /proc/sys/net/ipv6/conf/{}/autoconf".format(ifname))
        self.execute_shell_command("/scripts/apply_networking.py {}".format(" ".join(options)))
        
        try:
            utils.restore_mac(name, ifname, ifmode)
        except:
            pass
            
        if onbt == "yes":
            try:
                if interface_type != "lte":
                    utils.do_ifup(self, ifname)
                else:
                    if enable or enable=="true" or enable=="True":
                        utils.do_ifup(self, ifname)
            except subprocess.CalledProcessError as ex:
                self.logger.error('ifup command failed!')

        if dns:
            dnses = dns.split(',')
            dnsopts = ''.join(map(lambda dns_srv: ' '+ dns_srv, dnses))            
            subprocess.call('/usr/bin/resolvectl -4 dns {} {}'.format(name, dnsopts), shell=True)
        if dns6:
            dns6es = dns6.split(',')
            dns6opts = ''.join(map(lambda dns_srv: ' '+ dns_srv, dns6es))            
            subprocess.call('/usr/bin/resolvectl -6 dns {} {}'.format(ifname, dns6opts), shell=True)

        if ipv4_prefix or ipv6_prefix:
            vpputils.handle_vpp_interface_addresses(self, ifname, ipv4_prefix, ipv6_prefix)
            if self.get_kubernetes_bind_interface() == name:
                # We prioritize IPv6
                ip_prefix = ipv6_prefix or ipv4_prefix
                ip = ip_prefix.split('/')[0]
                #some legacy code left here, lets notify it also
                subprocess.Popen(['/scripts/set_openstack_address.sh', f'{ip}'])
                subprocess.Popen(['/scripts/handle_k8s_ip_change.sh', f'{ip}'])

        utils.config_smart_protocol_passthrough(self, name, config, change_set)

        # Reconnect all ipsecvpn that have local IP equal to IP of adaptor that is going to be down. Connect.
        for ipsec_iface in ipsecvpn_reenable_list:
            ipsec.ipsecvpn_connect(self, ipsec_iface)

        ipsecvpn_reenable_list = []

    def __post_interface_config(self):
        if self.igb_recalc_needed:
            utils.recalc_igb_params(self)
            if self.initramfs_changed:
                self.logger.warning("Requesting system reboot.[initramfs change]")
                utils.request_tve_reboot()

    def __pre_interface_config(self):
        try:
            pass
        except:
            self.logger.debug("systemctl stop trace_host failed!")
            pass

    def apply_startup_config(self, change_set=None):
        if 'yes' == os.environ.get('RTCTL', ''):
            vlanutils.ENABLE_DRCTL=False
            vlanutils.ENABLE_RT_CTL=True

        config_indexed = self.get_interfaces(startup=True)
        vlanutils.init_mac_db(config_indexed)
        # nat_routing_flag = False
        for name, config in sorted(config_indexed.items()):
            if config['type'] == "ipsecvpn":
                self.ipsvpn_interfaces.append(ips.ipsec_vpn_if(name))

            self.__config_interface(name, config, change_set, startup=True)
            self.__vswitch_if_config(name, config, None, None)

        self.__post_interface_config()
        radvd.init(interfaces=self.radvd_ifs, logger=self.logger)
        if self.tracehost is not None:
            if (self.tracehost.load_initial() > 0):
                self.logger.error("Failed to load initial tracehost interfaces configuration")

        return True

    def apply_config(self, change_set=None, apply_startup=False):
        if apply_startup:
            return self.apply_startup_config(change_set)

        # Get current running
        config_indexed = self.get_interfaces()

        self.__pre_interface_config()
        interface_changes = {}
        for change in change_set:
            
            # reinit wireguard connection in order for DNS server on Concentrator to update client hostname
            if change['xpath'] == '/ucpe:config/system:system/hostname':
                for key, value in config_indexed.items():
                    if key.startswith('wg') and 'wireguard-concentrator' in value and value.get('enable'):
                        wg_if = next(( a for a in self.wg_interfaces if a.name == key), None)
                        if wg_if.name == key:
                            wg_if.enable = False
                            wg_if.enable = True
                        
            if not change["xpath"].startswith(self.namespace):
                continue

            if change['operation']=='CREATED' and change['new_value']=="":
                if change["xpath"].endswith("/dns") or change["xpath"].endswith("/gateway") or change["xpath"].endswith("/trace-host") or change["xpath"].endswith("/apn"):
                        continue

            name = str(change['xpath'])
            name = name[name.find("interface[name='") + len("interface[name='"):]
            name = name[:name.find("]") - 1]
            if name not in interface_changes:
                interface_changes[name] = {}

            interface_changes[name][change['xpath'][(change['xpath'].find(']/') + len(']/')):]] = change

        try:
            down_interfaces = []
            # ifdown the interfaces which metrics have been modified
            for name, xpath_dict in interface_changes.items():
                if name in config_indexed and self.tracehost is not None and ('data' in name or 'lte' in name or 'wifi' in name or 'ppp' in name or 'mgmt' in name):
                    self.tracehost.del_config(name)
                if "metric" in xpath_dict:
                    sys_name = utils.get_ifname(self,name)
                    self.execute_shell_command(f"ifdown  {sys_name} || true")
                    down_interfaces.append(sys_name)

            # Go over modified interfaces
            for name, xpath_dict in interface_changes.items():
                if name in config_indexed:
                    self.__config_interface(name, config_indexed[name], xpath_dict)
                    if 'vswitch' in xpath_dict.keys():
                        self.__vswitch_if_config(name, config_indexed[name], xpath_dict, interface_changes)
                else:
                    if 'vti' in name:
                        self.ipsvpn_interfaces.append(ips.ipsec_vpn_if(name))
                    # TODO: remove after multi VLAN implementation. keep for reference for now
                    # if 'ppp' in name:
                    #    if 'ethernet-interface' in xpath_dict.keys():
                    #        oper = xpath_dict.get('ethernet-interface').get('operation', None)
                    #        if oper == 'MODIFIED' or oper == 'CREATED':
                    #            ethif = xpath_dict.get('ethernet-interface').get('new_value', None)
                    #            if ethif in interface_changes.keys():
                    #                if 'vlan' in interface_changes.get(ethif, None).keys():
                    #                    xpath_dict.update({'eth-if-vlan': interface_changes.get(ethif).get('vlan')})
                    if 'vxlan' in name:
                        if 'local/vtep-if' in xpath_dict.keys():
                            oper = xpath_dict.get('local/vtep-if').get('operation', None)
                            if oper == 'MODIFIED' or oper == 'CREATED':
                                ethif = xpath_dict.get('local/vtep-if').get('new_value', None)
                                if ethif in interface_changes.keys():
                                    if 'ipv4-prefix' in interface_changes.get(ethif, None).keys():
                                        xpath_dict.update({'local/lvtep-if-ipv4-prefix': interface_changes.get(ethif)
                                                          .get('ipv4-prefix')})

                    self.__config_interface(name, None, xpath_dict)

            # ifup the interfaces which metrics have been changed
            if down_interfaces:
                for oneif in down_interfaces:
                    utils.do_ifup(self, oneif)
                del down_interfaces

        except IfUpError as ex:
            self.logger.error(ex)
            return False

        radvd.init(interfaces=self.radvd_ifs, logger=self.logger)

        # All done. Save the config!
        return True

    def post_apply_config(self, change_set):
        if not change_set:
            return

        trace_host_interfaces = []

        for change in change_set:
            name = str(change['xpath'])
            name = name[name.find("interface[name='") + len("interface[name='"):]
            name = name[:name.find("]") - 1]
            iftype = ""
            if name:
                iftype = self.get_interface_param(name, 'type')

            if iftype == 'ethernet' and change['xpath'].find('mode') > -1:
                self.logger.warning("Interface mode changed to {}. Requesting system reboot.".format(change['new_value']))
                utils.request_tve_reboot()

            if iftype in ['ethernet', 'lte', 'wifi', 'pppoe', 'vlan'] and change['xpath'].find('ip-passthrough') > -1:
                if (('old_value' in change and 'new_value' in change and (str(change['old_value']) != str(change['new_value']))) or
                    ('old_value' not in change and 'new_value' in change and 'true' in str(change['new_value']).lower()) or
                    ('old_value' in change and 'new_value' not in change and 'true' in str(change['old_value']).lower())):
                    self.logger.warning("Interface {} - IP Passthrough changed to {}. Requesting system reboot.".format(name, change['new_value']))
                    utils.request_tve_reboot()
            if iftype in ['ethernet', 'lte', 'wifi', 'pppoe', 'vlan']:
                trace_host_interfaces.append(name)

        wifi.post_apply_config_wifi(self, change_set)
        self.__post_interface_config()
        if trace_host_interfaces and self.tracehost is not None:
            for ifname in trace_host_interfaces:
                if (self.tracehost.load_config(ifname) > 0):
                    self.logger.error("Failed to load interface {} config" .format(ifname))

    def status_get_cb(self, xpath, private_data):
        data = {'ucpe:status': {'interfaces:interface': []}}
        ipsvpn_fields = ['conname','state','established','rekey-time','vips','nat-local','nat-any','vips','local','remote','local-id',
                         'remote-id','version','integ-alg','encr-alg','encr-ksize','dh-group']
        ipsvpn_ch_fields = ['chname','state','bytes-in','bytes-out','pkt-in','pkt-out','mode','proto','local-ts','remote-ts','integ-alg',
                            'encr-alg','encr-ksize','dh-group']

        for vpn in self.ipsvpn_interfaces:
            sa = vpn.get_sa(all_child=True)
            iface_data = {'name': str(vpn.name)}

            if sa is not None:
                for field in ipsvpn_fields:
                    iface_data[str(field)] = str(sa.get(field, ""))

                chs = sa['child']
                if chs is not None:
                    iface_data['child'] = []
                    for ch in chs:
                        ch_data = {'chname': str(ch.get('chname'))}
                        for ch_field in ipsvpn_ch_fields:
                            ch_data[str(ch_field)] = str(ch.get(ch_field, ""))
                        iface_data['child'].append(ch_data)
            data['ucpe:status']['interfaces:interface'].append(iface_data)

        if len(data['ucpe:status']['interfaces:interface']) == 0:
            return None
        return data


    # Client helper functions

    def get_network_mode(self):
        return self.get_item("/ucpe:config/system:system/network-mode")

    def get_kubernetes_bind_interface(self):
        return self.get_item("/ucpe:config/kubernetes:kubernetes/bind-interface")

    def set_interface(self, name):
        self.set_item("{}[name='{}']".format(self.namespace, name), name)

    def get_interface(self, name):
        return self.get_item("{}[name='{}']".format(self.namespace, name))

    def get_interfaces(self, startup=False):
        config_indexed = {}
        sys_conf = self.get_startup_config(full_path=False) if startup else self.get_running_config(full_path=False)
        for interface in sys_conf:
            config_indexed[interface['name']] = interface
        return config_indexed

    def unset_interface(self, name):
        self.delete_item("{}[name='{}']".format(self.namespace, name))

    def set_interface_param(self, name, param, value):
        self.set_item("{}[name='{}']/{}".format(self.namespace, name, param), value)

    def get_interface_param(self, name, param, default=None):
        try:
            for interface in self.get_running_config(full_path=False):
                if interface['name'] == name and param in interface:
                    return self.get_item("{}[name='{}']/{}".format(self.namespace, name, param))
        except:
            return default

    def unset_interface_param(self, name, param):
        self.delete_item("{}[name='{}']/{}".format(self.namespace, name, param))

    def set_nat_wan_interface(self, nat_name, name):
        self.set_item("{}[name='{}']/wan-interface[name='{}']".format(self.namespace, nat_name, name), name)

    def get_nat_wan_interfaces(self, name):
        config_indexed = {}
        for interface in self.get_running_config(full_path=False):
            if interface['name'] == name:
                if 'wan-interface' in interface:
                    for wan in interface['wan-interface']:
                        config_indexed[wan['name']] = wan
        return config_indexed

    def unset_nat_wan_interface(self, nat_name, name):
        self.delete_item("{}[name='{}']/wan-interface[name='{}']".format(self.namespace, nat_name, name))

    def unset_nat_wan_interfaces(self, nat_name):
        self.delete_item("{}[name='{}']/wan-interface".format(self.namespace, nat_name))

    def set_concentrator_endpoint(self, ifname, value):
        self.set_item("{}[name='{}']/ovpn-cfg/concentrator-endpoint".format(self.namespace, ifname), value)

    def get_concentrator_endpoint(self, ifname):
        return self.get_item("{}[name='{}']/ovpn-cfg/concentrator-endpoint".format(self.namespace, ifname))

    def set_concentrator_port(self, ifname, value):
        self.set_item("{}[name='{}']/ovpn-cfg/concentrator-port".format(self.namespace, ifname), value)

    def get_concentrator_port(self, ifname):
        return self.get_item("{}[name='{}']/ovpn-cfg/concentrator-port".format(self.namespace, ifname))

    def get_ippass_param(self, ifname):
        return self.get_item("{}[name='{}']/ip-passthrough".format(self.namespace, ifname))

    def set_ippass_param(self, ifname, value):
        return self.set_item("{}[name='{}']/ip-passthrough".format(self.namespace, ifname), value)

    # Define get, set and unset methods for IKE1 local round parameters
    def unset_ike1_round(self, ifname, type, id):
        self.delete_item("{}[name='{}']/ike-phase-1/{}/rounds[id='{}']".format(self.namespace, ifname, type, id))

    def set_ike1_round(self, ifname, type, id):
        self.set_item("{}[name='{}']/ike-phase-1/{}/rounds[id='{}']".format(self.namespace, ifname, type, id), id)

    def set_ike1_round_param(self, ifname, type, id, param, value):
        self.set_item("{}[name='{}']/ike-phase-1/{}/rounds[id='{}']/{}".format(self.namespace, ifname, type, id, param), value)

    def unset_ike1_round_param(self, ifname, type, id, param):
        self.delete_item("{}[name='{}']/ike-phase-1/{}/rounds[id='{}']/{}".format(self.namespace, ifname, type, id, param))

    def unset_ike1_local_round(self, ifname, id):
        self.delete_item("{}[name='{}']/ike-phase-1/local/rounds[id='{}']".format(self.namespace, ifname, id))

    def unset_ike1_local_rounds(self, ifname):
        self.delete_item("{}[name='{}']/ike-phase-1/local/rounds".format(self.namespace, ifname))

    def set_ike1_local_round(self, ifname, id):
        self.set_item("{}[name='{}']/ike-phase-1/local/rounds[id='{}']".format(self.namespace, ifname, id), id)

    def get_ike1_local_rounds(self, name):
        local_rounds_indexed = {}
        for interface in self.get_running_config(full_path=False):
            if interface['name'] == name:
                if 'ike-phase-1' in interface:
                    for config in interface['ike-phase-1']:
                        if 'local' in config and 'rounds' in interface['ike-phase-1']['local']:
                            local_rounds_indexed = interface['ike-phase-1']['local']['rounds']
        return local_rounds_indexed

    def set_ike1_local_round_param(self, ifname, id, param, value):
        self.set_item("{}[name='{}']/ike-phase-1/local/rounds[id='{}']/{}".format(self.namespace, ifname, id, param), value)

    # Define get, set and unset methods for IKE1 remote round parameters
    def unset_ike1_remote_round(self, ifname, id):
        self.delete_item("{}[name='{}']/ike-phase-1/remote/rounds[id='{}']".format(self.namespace, ifname, id))

    def unset_ike1_remote_rounds(self, ifname):
        self.delete_item("{}[name='{}']/ike-phase-1/remote/rounds".format(self.namespace, ifname))

    def set_ike1_remote_round(self, ifname, id):
        self.set_item("{}[name='{}']/ike-phase-1/remote/rounds[id='{}']".format(self.namespace, ifname, id), id)

    def get_ike1_remote_rounds(self, name):
        remote_rounds_indexed = {}
        for interface in self.get_running_config(full_path=False):
            if interface['name'] == name:
                if 'ike-phase-1' in interface:
                    for config in interface['ike-phase-1']:
                        if 'remote' in config and 'rounds' in interface['ike-phase-1']['remote']:
                            remote_rounds_indexed = interface['ike-phase-1']['remote']['rounds']
        return remote_rounds_indexed

    def set_ike1_remote_round_param(self, ifname, id, param, value):
        self.set_item("{}[name='{}']/ike-phase-1/remote/rounds[id='{}']/{}".format(self.namespace, ifname, id, param), value)

    # Define get, set and unset methods for IKE1 Diffie-Hellman groups
    def set_dh_groups(self, phase, ifname, group):
        try:
            self.set_item("{}[name='{}']/{}/dh-groups[group='{}']".format(self.namespace, ifname, phase, group), group)
        except Exception as ex:
            self.logger.error('Set Diffie-Hellman groups "{group}" failed. Not valid group. {msg}'.format(group=group, msg=str(ex)))
            print('Set Diffie-Hellman groups "{group}" failed. Not valid group. {msg}'.format(group=group, msg=str(ex)))

    def set_ike1_dh_groups(self, ifname, group):
        try:
            self.set_item("{}[name='{}']/ike-phase-1/dh-groups[group='{}']".format(self.namespace, ifname, group), group)
        except Exception as ex:
            self.logger.error('Set Diffie-Hellman groups "{group}" failed. Not valid group. {msg}'.format(group=group, msg=str(ex)))
            print('Set Diffie-Hellman groups "{group}" failed. Not valid group. {msg}'.format(group=group, msg=str(ex)))

    def get_ike1_dh_groups(self, name):
        groups_indexed = {}
        for interface in self.get_running_config(full_path=False):
            if interface['name'] == name:
                if 'ike-phase-1' in interface:
                    for config in interface['ike-phase-1']:
                        if 'dh-groups' in config:
                            groups_indexed = interface['ike-phase-1']['dh-groups']
        return groups_indexed

    def unset_ike1_dh_group(self, ifname, group):
        self.delete_item("{}[name='{}']/ike-phase-1/dh-groups[group='{}']".format(self.namespace, ifname, group))

    def unset_dh_groups(self, phase, ifname):
        self.delete_item("{}[name='{}']/{}/dh-groups".format(self.namespace, ifname, phase))

    def unset_ike1_dh_groups(self, ifname):
        self.delete_item("{}[name='{}']/ike-phase-1/dh-groups".format(self.namespace, ifname))

    # Define get, set and unset methods for IKE1 proposals
    def set_ike1_proposals(self, ifname, id):
        self.set_item("{}[name='{}']/ike-phase-1/proposals[id='{}']".format(self.namespace, ifname, id), id)

    def get_ike1_proposal(self, ifname, id):
        return self.get_item("{}[name='{}']/ike-phase-1/proposals[id='{}']".format(self.namespace, ifname, id))

    def get_ike1_proposal_param(self, ifname, id, param):
        return self.get_item("{}[name='{}']/ike-phase-1/proposals[id='{}']/{}".format(self.namespace, ifname, id, param))

    def get_proposal_param(self, phase, ifname, id, param):
        return self.get_item("{}[name='{}']/{}/proposals[id='{}']/{}".format(self.namespace, ifname, phase, id, param))

    def set_proposal_param(self, phase, ifname, id, param, value):
        self.set_item("{}[name='{}']/{}/proposals[id='{}']/{}".format(self.namespace, ifname, phase, id, param), value)

    def get_ike1_proposals(self, name):
        proposals_indexed = {}
        for interface in self.get_running_config(full_path=False):
            if interface['name'] == name:
                if 'ike-phase-1' in interface:
                    for config in interface['ike-phase-1']:
                        if 'proposals' in config:
                            proposals_indexed = interface['ike-phase-1']['proposals']
        return proposals_indexed

    def unset_proposal(self, phase, ifname, id):
        self.delete_item("{}[name='{}']/{}/proposals[id='{}']".format(self.namespace, ifname, phase, id))

    def set_proposal(self, phase, ifname, id):
        self.set_item("{}[name='{}']/{}/proposals[id='{}']".format(self.namespace, ifname, phase, id), id)

    def unset_ike1_proposal(self, ifname, id):
        self.delete_item("{}[name='{}']/ike-phase-1/proposals[id='{}']".format(self.namespace, ifname, id))

    def unset_ike1_proposals(self, ifname):
        self.delete_item("{}[name='{}']/ike-phase-1/proposals".format(self.namespace, ifname))

    def set_ike1_proposals_param(self, ifname, id, param, value):
        self.set_item("{}[name='{}']/ike-phase-1/proposals[id='{}']/{}".format(self.namespace, ifname, id, param), value)

    # Define get, set and unset methods for IKE2 local and remote traffic selectors
    def set_ike2_local_ts(self, ifname, value):
        self.set_item("{}[name='{}']/ike-phase-2/local-ts".format(self.namespace, ifname), value)

    def get_ike2_local_ts(self, ifname):
        return self.get_item("{}[name='{}']/ike-phase-2/local-ts".format(self.namespace, ifname))

    def unset_ike2_local_ts(self, ifname):
        self.delete_item("{}[name='{}']/ike-phase-2/local-ts".format(self.namespace, ifname))

    def set_ike2_remote_ts(self, ifname, value):
        self.set_item("{}[name='{}']/ike-phase-2/remote-ts".format(self.namespace, ifname), value)

    def get_ike2_remote_ts(self, ifname):
        return self.get_item("{}[name='{}']/ike-phase-2/remote-ts".format(self.namespace, ifname))

    def unset_ike2_remote_ts(self, ifname):
        self.delete_item("{}[name='{}']/ike-phase-2/remote-ts".format(self.namespace, ifname))

    # Define get, set and unset methods for IKE2 Diffie-Hellman groups
    def set_ike2_dh_groups(self, ifname, group):
        try:
            self.set_item("{}[name='{}']/ike-phase-2/dh-groups[group='{}']".format(self.namespace, ifname, group), group)
        except Exception as ex:
            self.logger.error('Set Diffie-Hellman groups "{group}" failed. Not valid group. {msg}'.format(group=group, msg=str(ex)))
            print('Set Diffie-Hellman groups "{group}" failed. Not valid group. {msg}'.format(group=group, msg=str(ex)))

    def get_ike2_dh_groups(self, name):
        groups_indexed = {}
        for interface in self.get_running_config(full_path=False):
            if interface['name'] == name:
                if 'ike-phase-2' in interface:
                    for config in interface['ike-phase-2']:
                        if 'dh-groups' in config:
                            groups_indexed = interface['ike-phase-2']['dh-groups']
        return groups_indexed

    def unset_ike2_dh_group(self, ifname, group):
        self.delete_item("{}[name='{}']/ike-phase-2/dh-groups[group='{}']".format(self.namespace, ifname, group))

    def unset_ike2_dh_groups(self, ifname):
        self.delete_item("{}[name='{}']/ike-phase-2/dh-groups".format(self.namespace, ifname))

    # Define get, set and unset methods for IKE2 proposals
    def set_ike2_proposals(self, ifname, id):
        self.set_item("{}[name='{}']/ike-phase-2/proposals[id='{}']".format(self.namespace, ifname, id), id)

    def get_ike2_proposal(self, ifname, id):
        return self.get_item("{}[name='{}']/ike-phase-2/proposals[id='{}']".format(self.namespace, ifname, id))

    def get_ike2_proposal_param(self, ifname, id, param):
        return self.get_item("{}[name='{}']/ike-phase-2/proposals[id='{}']/{}".format(self.namespace, ifname, id, param))

    def get_ike2_proposals(self, name):
        proposals_indexed = {}
        for interface in self.get_running_config(full_path=False):
            if interface['name'] == name:
                if 'ike-phase-2' in interface:
                    for config in interface['ike-phase-2']:
                        if 'proposals' in config:
                            proposals_indexed = interface['ike-phase-2']['proposals']
        return proposals_indexed

    def unset_ike2_proposal(self, ifname, id):
        self.delete_item("{}[name='{}']/ike-phase-2/proposals[id='{}']".format(self.namespace, ifname, id))

    def unset_ike2_proposals(self, ifname):
        self.delete_item("{}[name='{}']/ike-phase-2/proposals".format(self.namespace, ifname))

    def set_ike2_proposals_param(self, ifname, id, param, value):
        self.set_item("{}[name='{}']/ike-phase-2/proposals[id='{}']/{}".format(self.namespace, ifname, id, param), value)

    def get_config_mode(self, ifname):
        return self.get_item("{}[name='{}']/config-mode".format(self.namespace, ifname))

    def set_config_mode(self, ifname, value):
        self.set_item("{}[name='{}']/config-mode".format(self.namespace, ifname), value)

    def get_install_vip(self, ifname):
        return self.get_item("{}[name='{}']/install-virtual-ip".format(self.namespace, ifname))

    def set_install_vip(self, ifname, value):
        self.set_item("{}[name='{}']/install-virtual-ip".format(self.namespace, ifname), value)

    def unset_install_vip(self, ifname):
        self.delete_item("{}[name='{}']/install-virtual-ip".format(self.namespace, ifname))

    def get_vip(self, ifname):
        return self.get_item("{}[name='{}']/virtual-ip".format(self.namespace, ifname))

    def set_vip(self, ifname, value):
        self.set_item("{}[name='{}']/virtual-ip".format(self.namespace, ifname), value)

    def unset_vip(self, ifname):
        self.delete_item("{}[name='{}']/virtual-ip".format(self.namespace, ifname))

    def get_if_status(self, ifname):
        data = dict()
        xpath = "{}[name='{}']".format(self.status_namespace, ifname)

        values = self.session.get_items(xpath + "//*")
        if not values:
            return data
        for i in range(values.val_cnt()):
            val = self.get_data_value(values.val(i))
            data[values.val(i).xpath()] = val

        return data

    def set_interface_wifi_securty_param(self, name, param, value):
        self.set_item("{}[name='{}']/security/{}".format(self.namespace, name, param), value)

    def get_interface_wifi_securty_param(self, param):
        return self.get_item("/ucpe:config/interfaces:interface[name='wifi0']/security/{}".format(param))

    def unset_interface_wifi_securty_param(self, name, param):
        self.delete_item("{}[name='{}']/security/{}".format(self.namespace, name, param))

    def set_interface_wifi_radio_param(self, name, param, value):
        self.set_item("{}[name='{}']/radio-settings/{}".format(self.namespace, name, param), value)

    def get_interface_wifi_radio_param(self, name, param, default=None):
        try:
            for interface in self.get_running_config(full_path=False):
                if interface['name'] == name and param in interface:
                    return self.get_item("{}[name='{}']/radio-settings/{}".format(self.namespace, name, param))
        except:
            return default

    def unset_interface_wifi_radio_param(self, name, param):
        self.delete_item("{}[name='{}']/radio-settings/{}".format(self.namespace, name, param))

    def set_wifi_wan_interface(self, wifi_name, name):
        self.set_item("{}[name='{}']/wifi-wan-interface[name='{}']".format(self.namespace, wifi_name, name), name)

    def get_wifi_wan_interfaces(self, name):
        config_indexed = {}
        for interface in self.get_running_config(full_path=False):
            if interface['name'] == name:
                if 'wifi-wan-interfacee' in interface:
                    for wan in interface['wifi-wan-interface']:
                        config_indexed[wan['name']] = wan
        return config_indexed

    def unset_wifi_wan_interface(self, wifi_name, name):
        self.delete_item("{}[name='{}']/wifi-wan-interface[name='{}']".format(self.namespace, wifi_name, name))

    def unset_wifi_wan_interfaces(self, wifi_name):
        self.delete_item("{}[name='{}']/wifi-wan-interface".format(self.namespace, wifi_name))

    def set_mac_allow(self, ifname, mac):
        self.set_item("{}[name='{}']/security/mac-allow[mac='{}']".format(self.namespace, ifname, mac), mac)

    def get_mac_allow(self, ifname, mac):
        return self.get_item("{}[name='{}']/security/mac-allow[mac='{}']".format(self.namespace, ifname, mac))

    def get_mac_allow_list(self, name):
        macs = []
        for interface in self.get_running_config(full_path=False):
            if interface['name'] == name:
                if 'mac-allow' in interface:
                    macs = interface['mac-allow']
                break
        return macs

    def unset_mac_allow(self, ifname, mac):
        self.delete_item("{}[name='{}']/security/mac-allow[mac='{}']".format(self.namespace, ifname, mac))

    def unset_mac_allow_list(self, ifname):
        self.delete_item("{}[name='{}']/security/mac-allow".format(self.namespace, ifname))

    def set_mac_deny(self, ifname, mac):
        self.set_item("{}[name='{}']/security/mac-deny[mac='{}']".format(self.namespace, ifname, mac), mac)

    def get_mac_deny(self, ifname, mac):
        return self.get_item("{}[name='{}']/security/mac-deny[mac='{}']".format(self.namespace, ifname, mac))

    def get_mac_deny_list(self, name):
        macs = []
        for interface in self.get_running_config(full_path=False):
            if interface['name'] == name:
                if 'mac-deny' in interface:
                    macs = interface['mac-deny']
                break
        return macs

    def unset_mac_deny(self, ifname, mac):
        self.delete_item("{}[name='{}']/security/mac-deny[mac='{}']".format(self.namespace, ifname, mac))

    def unset_mac_deny_list(self, ifname):
        self.delete_item("{}[name='{}']/security/mac-deny".format(self.namespace, ifname))

    def get_container_param(self, ifname, container, param, default=None):
        try:
            interface = next((intf for intf in self.get_running_config(full_path=False) if intf['name'] == ifname), None)
            if interface is None:
                return default
            else:
                return self.get_item("{}[name='{}']/{}/{}".format(self.namespace, ifname, container, param))
        except:
            return default

    def set_container_param(self, ifname, container, param, value):
        self.set_item("{}[name='{}']/{}/{}".format(self.namespace, ifname, container, param), value)

    def unset_container_param(self, ifname, container, param):
        self.delete_item("{}[name='{}']/{}/{}".format(self.namespace, ifname, container, param))

    def unset_tracehost_probes(self, ifname):
        self.delete_item("{}[name='{}']/tracehost/probes".format(self.namespace, ifname))

    def unset_tracehost_probe(self, ifname, id):
        self.delete_item("{}[name='{}']/tracehost/probes[id='{}']".format(self.namespace, ifname, id))

    def set_tracehost_probe(self, ifname, id):
        self.set_item("{}[name='{}']/tracehost/probes[id='{}']".format(self.namespace, ifname, id), id)

    def set_tracehost_probe_param(self, ifname, id, param, value):
        self.set_item("{}[name='{}']/tracehost/probes[id='{}']/{}".format(self.namespace, ifname, id, param), value)

    def unset_tracehost_probe_param(self, ifname, id, param):
        self.delete_item("{}[name='{}']/tracehost/probes[id='{}']/{}".format(self.namespace, ifname, id, param))

    def set_smart_pp_param(self, ifname, protocol, param, value):
        self.set_item("{}[name='{}']/smart-protocol-passthrough[protocol='{}']/{}".format(self.namespace, ifname, protocol, param), value)

    def del_smart_pp(self, ifname, protocol):
        self.delete_item("{}[name='{}']/smart-protocol-passthrough[protocol='{}']".format(self.namespace, ifname, protocol))

    def unset_wireguard_peers(self, ifname):
        self.delete_item("{}[name='{}']/peers".format(self.namespace, ifname))

    def unset_wireguard_peer(self, ifname, endpoint):
        self.delete_item("{}[name='{}']/peers[peer-endpoint='{}']".format(self.namespace, ifname, endpoint))

    def set_wireguard_peer(self, ifname, endpoint):
        self.set_item("{}[name='{}']/peers[peer-endpoint='{}']".format(self.namespace, ifname, endpoint), endpoint)

    def set_wireguard_peer_param(self, ifname, endpoint, param, value):
        self.set_item("{}[name='{}']/peers[peer-endpoint='{}']/{}".format(self.namespace, ifname, endpoint, param), value)

    def unset_wireguard_peer_param(self, ifname, endpoint, param):
        self.delete_item("{}[name='{}']/peers[peer-endpoint='{}']/{}".format(self.namespace, ifname, endpoint, param))

    def get_wireguard_peer_params(self, name, param, default=None):
        try:
            for interface in self.get_running_config(full_path=False):
                if interface['name'] == name and 'peers' in interface:
                    return self.get_item("{}[name='{}']/peers/{}".format('/ucpe:config/interfaces:interface', name, param))
        except:
            return default
        
    def get_wireguard_peer_all_params(self, name, default=None):
        try:
            for interface in self.get_running_config(full_path=False):
                if interface['name'] == name and 'peers' in interface:
                    return self.session.get_data("{}[name='{}']/peers".format('/ucpe:config/interfaces:interface', name))
        except:
            return default

    def set_wireguard_concentrator_param(self, ifname, param, value):
        self.set_item("{}[name='{}']/wireguard-concentrator/{}".format(self.namespace, ifname, param), value)

    def unset_wireguard_concentrator_param(self, ifname, param):
        self.delete_item("{}[name='{}']/wireguard-concentrator/{}".format(self.namespace, ifname, param))
        
    def del_wireguard_concentrator(self, name):
        self.delete_item("{}[name='{}']/wireguard-concentrator".format(self.namespace, name))

    def get_wireguard_concentrator_param(self, name, param, default=None):
        try:
            for interface in self.get_running_config(full_path=False):
                if interface['name'] == name and 'wireguard-concentrator' in interface:
                    return self.get_item("{}[name='{}']/wireguard-concentrator/{}".format(self.namespace, name, param))
        except:
            return default
        
    def get_wireguard_concentrator_params(self, name, default=None):
        try:
            for interface in self.get_running_config(full_path=False):
                if interface['name'] == name and 'wireguard-concentrator' in interface:
                    return self.session.get_data("{}[name='{}']/wireguard-concentrator".format(self.namespace, name))
        except:
            return default

    def get_tracehost_probes(self, name):
        config_indexed = {}
        for interface in self.get_running_config(full_path=False):
            if interface['name'] == name:
                if 'tracehost' in interface:
                    if 'probes' in interface['tracehost']:
                        for probe in interface['tracehost']['probes']:
                            config_indexed[probe['id']] = probe
        return config_indexed

    def get_vxlan_rvtep(self, ifname):
        rvteps = []
        vals = self.session.get_items("{}[name='{}']/remote/vtep-addrs//*".format(self.namespace, ifname))
        if vals:
            for val in vals:
                if val:
                    rvteps.append(self.get_data_value(val))
        return rvteps

    def add_vxlan_rvtep(self, ifname, address):
        # get current list of vteps
        cvteps = self.get_vxlan_rvtep(ifname)
        if address in cvteps:
            # allready in config
            return

        return self.set_item("{}[name='{}']/remote/vtep-addrs[vtep-addr='{}']".format(self.namespace, ifname, address), address)

    def del_vxlan_rvtep(self, ifname, address):
        # get current list of vteps
        cvteps = self.get_vxlan_rvtep(ifname)
        if address not in cvteps:
            # allready in config
            return

        return self.delete_item("{}[name='{}']/remote/vtep-addrs[vtep-addr='{}']".format(self.namespace, ifname, address))

    def del_vxlan_rvteps(self, ifname):
        return self.delete_item("{}[name='{}']/remote/vtep-addrs".format(self.namespace, ifname))

    def tag_interface(self, iface, vlan):
        return utils.tag_interface(iface, vlan)

    def get_wireguard_peer_all_params(self, name, default=None):
        try:
            for interface in self.get_running_config(full_path=False):
                if interface['name'] == name and 'peers' in interface:
                    return self.session.get_data("{}[name='{}']/peers".format('/ucpe:config/interfaces:interface', name))
        except:
            return default
